module objectIOwarmup
import StdEnv, StdIO

//////////////////////////////////////////////////
//           Clean Warm-up : ObjectIO

/*
 goal of ObjectIO : work with interactive objects
 interactive objects : 
  - windows, menus etc.
  - represented as algebraic data types;
      the names of type and data constructors coincide
  - all interactive objects are in the I/O state,
    which is part of the process state
*/

//////////////////////////////////////////////////
// A simple dialogue window

/*
Start :: *World -> *World
 // the unique world has every information about the
 // running environment
Start world	= startIO NDI Void init [] world
 // startIO initialises an IO environment for us
 //   NDI  : our simple example uses no document
 //   Void : we could have a local state
where
 init pst
  # ( error, pst ) = openDialog Void hello pst
      // beginning of the life cycle of the object
      // # : the rules are used in sequence
	  // hello : describes the contents of the dialogue window
  | error <> NoError = closeProcess pst
      // closeProcess is a callback function of ObjectIO
      // callbacks are called upon an event
      //   they perform state transitions
  | otherwise      = pst
	  // if the window could not be opened,
	  // we return the previous process state

 hello
  = Dialog
     "Dialog title"
     ( TextControl "Example text" [] )
        // window contents
     [ WindowClose ( noLS closeProcess ) ]
        // [ ... ] : window attribute list
        // noLS    : conversion to appropriate type (no local state)
*/

//////////////////////////////////////////////////
// A dialogue with a counter

/*
Start world
 # ( id, world ) = openId world
      // an identifier to reference the counter
 = startIO NDI
           Void
           ( snd o openDialog Void ( dialogue id ) )
              // composing snd : we expect no error
           [ProcessClose closeProcess]
			  // ProcessClose is a process attribute
           world

dialogue id
	// the id is passed to the dialogue
	// as one of its controls is referenced through it
 = Dialog "Counter" ( counter id )
    [ WindowClose    ( noLS closeProcess )
	, WindowViewSize { w = 150, h = 100 }
	]

counter id
 = { newLS  = count0  // local state : current value
   , newDef =
      LayoutControl
       (   EditControl
             ( toString count0 )
             ( ContentWidth "000" )
             1           // number of lines
             [ ControlSelectState Unable
             , ControlId id
				// passed a long way...
             ]
       :+: ButtonControl // :+: is the universal glue
             "-"
             [ ControlFunction ( count (-1) ) ]
             	// arguments of ButtonControls: text, attribs
				// ControlFunction: callback on action (button press)
       :+: ButtonControl
             "+"
             [ ControlFunction ( count 1 ) ]
       )
       []
   }
where
  count0 = 0
	// it is good practice to use named constants
	// instead of writing the numbers in the code itself ("magic numbers")

  count difference ( count, pst )
           // callback for state transition
           // count : local state
           // pst   : the process state
    = ( count + difference      // new local state
      , appPIO
         ( setControlText
            id
            ( toString ( count + difference ) )
         )
         pst
      )
*/

//////////////////////////////////////////////////
// Exercise #1

/*

Create two counters (as before, with two buttons and an EditControl to display the value),
one more EditControl and one button.
When this button is pressed,
print the sum of the counters into the EditControl.
*/

//////////////////////////////////////////////////
// Making menus

/*
Start world
 = startIO SDI Void (snd o (openMenu Void menu))
           [ProcessClose closeProcess] world
where
 menu =
  Menu
   "&File"
   ( MenuItem "E&xit"
              [ MenuShortKey 'x'
              , MenuFunction ( noLS closeProcess )
              ]
		// &x : the letter is underlined
		// MenuShortKey : pressing x actually does something
		//                currently it closes the window
   )
   []
*/

//////////////////////////////////////////////////
// Drawing to a window

/*
  
There are two ways to do it.

1. Indirect drawing
	- you edit an underlying document
	- a Look displays the document
	- this is the approach taken here,
	  although extremely simplified

2. Direct drawing to the window

*/

// You may have to increase the maximum heap size.
// Set "Maximum Heap Size" in Project -> Project Options... to 4000K.

/*
Start w
 # ( f, w )    = selectInputFile w
 | isNothing f = w
 # ( b, w )	   = openBitmap ( fromJust f ) w
 | isNothing b = w
 # bitmap      = fromJust b
   window
    = Window
       "Bitmap" NilLS
       [ WindowViewSize (getBitmapSize bitmap)
       , WindowLook True (\_ _ -> drawAt zero bitmap)
           // the second argument updates the look
       , WindowClose (noLS closeProcess)
       ]
 = startIO SDI Void
           ( snd o openWindow Void window )
	       [ProcessClose closeProcess] w
*/

//////////////////////////////////////////////////
// Exercise #2

/*
Create a window that has a menu item besides Exit
that can open a bitmap from a menu and draw it.

Hints:
- windows have a WindowId attribute
- the file operations also work on pst
- an easy way to change window attribute XYZ is the following
  (you should provide the id of the window and the parameters)

  		appPIO (setXYZ window_id parameters) pst

*/

//////////////////////////////////////////////////
// A simple  animation

/*
:: TimerState = { radius :: Int }

Start world	= circles ( openId world )

circles ( wid, world )
 = startIO SDI Void
     ( snd o seqList [ openWindow Void      wdef
                     , openTimer  initState tdef
                     ]
     ) [ProcessClose closeProcess] world
where
 size = 350
 area = { corner1 = { x = ~size, y = ~size }
        , corner2 = { x =  size, y =  size }
        }
 wdef = Window "Circles" NilLS
          [ WindowId         wid
          , WindowViewDomain area
          ]

 tdef = Timer ( ticksPerSecond / 50 ) NilLS 
              [ TimerFunction action ]
    // the action is invoked 20 times a second
 initState = { radius = 1 }

 circ r = { oval_rx = r, oval_ry = r }
   // helper function to (un)draw a circle
 action _ ( lst =: { radius }, pst )
  | radius < size
    # lst = { lst & radius = radius + 1 }
    # pst = appPIO
             ( appWindowPicture wid
                ( (draw (circ radius)) o (undraw (circ (radius - 1))) )
             ) pst
    = ( lst, pst )
  | otherwise
    # pst = appPIO ( appWindowPicture
                       wid ( unfill area ) )
                   pst
    = ( initState, pst )
*/

//////////////////////////////////////////////////
// Exercise #3

/*

Create a menu for the previous program to control its speed.

Hints :
- timers' attribute is TimerId, of course
- only fixed speed values are needed
- the timer can be changed by
  (io is a component of pst, as in pst=:{io})
  (you have to supply the id of the timer and the new speed)

		setTimerInterval timer_id new_speed io

*/

//////////////////////////////////////////////////
// Thank you for your attention

/*
   Thanks go to the whole ObjectIO team.

   Some examples are based on the ObjectIO tutorial
     - introduction/hello
     - controls/counter
     - menus
     - timers/circles
*/
